/*
Copyright (C) 2003-2004 Douglas Thain and the University of Wisconsin
Copyright (C) 2005- The University of Notre Dame
This software is distributed under a BSD-style license.
See the file COPYING for details.
*/

#include "chirp_exec.h"
#include "chirp_protocol.h"

#include "stringtools.h"
#include "link.h"
#include "full_io.h"
#include "debug.h"
#include "macros.h"
#include "copy_stream.h"

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <sys/wait.h>

extern int enable_identity_boxing;
extern const char *chirp_server_path;

int chirp_exec( struct link * link, const char *subject, const char *infile, const char *outfile, const char *errfile, const char *path, const char *args )
{
	int i;
	int argc;
	char **argv;
	int maxfd = sysconf(_SC_OPEN_MAX);
	int fds[3];
	pid_t pid;
	char realargs[CHIRP_LINE_MAX];
	char dir[CHIRP_PATH_MAX];

	/* Determine the working directory of the program */
	string_dirname(path,dir);

	/* chop off the dummy leading argument, if any*/
	args = strchr(args,' ');
	if(args) {
		args++;
	} else {
		args="";
	}

	/* The identity box must be in the same place as the server */	
	if(enable_identity_boxing) {
		sprintf(realargs,"parrot_identity_box %s %s %s",subject,path,args);
	} else {
		sprintf(realargs,"%s %s",path,args);
	}

	debug(D_PROCESS,"running: %s",realargs);

	if(access(path,X_OK)!=0) {
		debug(D_PROCESS,"%s can't be executed: %s\n",path,strerror(errno));
		return -1;
	}

	debug(D_PROCESS,"stdin: %s\n",infile);
	fds[0] = open(infile,O_RDWR,0);
	if(fds[0]<0) {
		debug(D_PROCESS,"couldn't open %s: %s",infile,strerror(errno));
		return -1;
	}

	debug(D_PROCESS,"stdout: %s\n",outfile);
	fds[1] = open(outfile,O_WRONLY|O_TRUNC|O_CREAT,0777);
	if(fds[1]<0) {
		debug(D_PROCESS,"couldn't open %s: %s",outfile,strerror(errno));
		close(fds[0]);
		return -1;
	}

	debug(D_PROCESS,"stderr: %s\n",errfile);
	if(!strcmp(errfile,outfile)) {
		fds[2] = dup(fds[1]);
	} else {
		fds[2] = open(errfile,O_WRONLY|O_TRUNC|O_CREAT,0777);
		if(fds[2]<0) {
			debug(D_PROCESS,"couldn't open %s: %s",errfile,strerror(errno));
			close(fds[0]);
			close(fds[1]);
			return -1;
		}
	}

	/* Flush all stdio buffers so that the child does not inherit them. */
	fflush(0);

	pid = fork();
	if(pid==0) {
		int result;

		/* Run in a new process group.  This is necessary so that the top-level chirp_server process doesn't accidentally wait for this sub-job in waitpid(-1) */
#ifdef CCTOOLS_OPSYS_DARWIN
		setpgrp(getpid(),getpid());
#else
		setpgrp();
#endif
		/* Put our standard input, output, and error in the expected fds */
		dup2(fds[0],0);
		dup2(fds[1],1);
		dup2(fds[2],2);

		/* Close all other fds so that this process doesn't muck with the parent's files */
		for(i=3;i<maxfd;i++) close(i);

		/* Split the args into pieces and execute the program */
		string_split((char*)realargs,&argc,&argv);

		/* Move to the dir containing the program */
		result = chdir(dir);
		if(result!=0) {
			printf("couldn't chdir %s: %s\n",dir,strerror(errno));
			_exit(1);
		}

		/* Execute the program */
		execvp(argv[0],argv);

		/* If execvp returns, the only possibility is an error. */
		printf("couldn't execute %s: %s\n",argv[0],strerror(errno));
		fflush(stdout);
		_exit(1);
	} else if(pid>0) {
		pid_t checkpid;
		int status;
		close(fds[0]);
		close(fds[1]);
		close(fds[2]);
		debug(D_PROCESS,"waiting for pid %d to complete...",pid);
		do {
			checkpid = waitpid(pid,&status,0);
			if(checkpid<0 && errno==ECHILD) {
				debug(D_PROCESS,"yikes: waitpid returned no children!\n");
				errno = ECHILD;
				return -1;
			}
		} while(checkpid!=pid);
		if(WIFEXITED(status)) {
			int exitcode = WEXITSTATUS(status);
			debug(D_PROCESS,"pid %d exited normally with status %d",pid,exitcode);
			return exitcode;
		} else {
			int signum = WTERMSIG(status);
			debug(D_PROCESS,"pid %d exited abnormally with status %d",pid,signum);
			return 256+signum;
		}
	} else {
		close(fds[0]);
		close(fds[1]);
		close(fds[2]);
		return -1;
	}

	return -1;
}
